/* Low-level statistical profiling support function.  Mostly POSIX.1 version.
   Copyright (C) 1996, 1997, 1998 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>

#ifndef SIGPROF

#include <sysdeps/generic/profil.c>

#else

static u_short *samples;
static size_t nsamples;
static size_t pc_offset;
static u_int pc_scale;
static int** pcsamples;
static u_int npcsamples;
static u_int npc = 0;
static int sampledata = 0;

u_int get_pcsample_size(void) {
  return npc;
}

static inline void
profil_count (void *pc)
{
  if (sampledata) {
    if (npc < npcsamples) {
      pcsamples[npc++] = (int*)pc;
    }
  }
  else {
    size_t i = (pc - pc_offset - (void *) 0) / 2;

    if (sizeof (unsigned long long int) > sizeof (size_t))
      i = (unsigned long long int) i * pc_scale / 65536;
    else
      i = i / 65536 * pc_scale + i % 65536 * pc_scale / 65536;
    if (i < nsamples)
      ++samples[i];
  }
}

/* Get the machine-dependent definition of `profil_counter', the signal
   handler for SIGPROF.  It calls `profil_count' (above) with the PC of the
   interrupted code.  */
#include "profil-counter.h"

/* Enable statistical profiling, writing samples of the PC into at most
   SIZE bytes of SAMPLE_BUFFER; every processor clock tick while profiling
   is enabled, the system examines the user PC and increments
   SAMPLE_BUFFER[((PC - OFFSET) / 2) * SCALE / 65536].  If SCALE is zero,
   disable profiling.  Returns zero on success, -1 on error.  */

int
__profil (u_short *sample_buffer, size_t size, size_t offset, u_int scale)
{
  struct itimerval timer;
  int proffreq = __profile_frequency();
#ifndef IS_IN_rtld
  static sighandler_t oact;
  static struct itimerval otimer;
# define oact_ptr &oact
# define otimer_ptr &otimer

  /* No profiling for sampling frequencies less than 1 HZ */
  if (proffreq < 1)
    return 0;

  if (sample_buffer == NULL)
    {
      /* Disable profiling.  */
      if (samples == NULL)
	/* Wasn't turned on.  */
	return 0;

      if (__setitimer (ITIMER_PROF, &otimer, NULL) < 0)
	return -1;
      samples = NULL;
      if (signal(SIGPROF, oact) == SIG_ERR)
	return -1;
      else
	return 0;
    }

 if (samples)
    {
      /* Was already turned on.  Restore old timer and signal handler
	 first.  */
      if (__setitimer (ITIMER_PROF, &otimer, NULL) < 0
	  || signal(SIGPROF, oact) == SIG_ERR)
	return -1;
    }
#else
 /* In ld.so profiling should never be disabled once it runs.  */
 //assert (sample_buffer != NULL);
# define oact_ptr NULL
# define otimer_ptr NULL
#endif

 if ((offset == 0) && (scale == 0)) {
   pcsamples = (int**)sample_buffer;
   npcsamples = size / sizeof *pcsamples;
   sampledata = 1;
 }
 else {
  samples = sample_buffer;
  nsamples = size / sizeof *samples;
  pc_offset = offset;
  pc_scale = scale;
  sampledata = 0;
 }

  if ((oact = signal(SIGPROF, (sighandler_t) &profil_counter)) == SIG_ERR)
    return -1;

  if (proffreq == 1) {
    timer.it_value.tv_sec = 1;
    timer.it_value.tv_usec = 0;
  } else {
    timer.it_value.tv_sec = 0;
    timer.it_value.tv_usec = 1000000 / proffreq;
  }
  timer.it_interval = timer.it_value;
  return __setitimer (ITIMER_PROF, &timer, otimer_ptr);
}

weak_alias (__profil, profil)

#endif
