/****************************************************************************/
/* ZEPTOOS:zepto-info */
/*     This file is part of ZeptoOS: The Small Linux for Big Computers.
 *     See www.mcs.anl.gov/zeptoos for more information.
 */
/* ZEPTOOS:zepto-info */
/* */
/* ZEPTOOS:zepto-fillin */
/*     $Id: init-zoid.c,v 1.18 2007/06/27 15:20:38 iskra Exp $
 *     ZeptoOS_Version: 1.2
 *     ZeptoOS_Heredity: FOSS_ORIG
 *     ZeptoOS_License: GPL
 */
/* ZEPTOOS:zepto-fillin */
/* */
/* ZEPTOOS:zepto-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and UChicago Argonne, LLC.  2004, 2005, 2006, 2007
 *      ZeptoOS License: GPL
 * 
 *      This software is free.  See the file ZeptoOS/misc/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* ZEPTOOS:zepto-gpl */
/****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>
#include <unistd.h>

#include <bgllockbox.h>
#include <bglmemmap.h>
#include <rts.h>

#include "bgl.h"
#include "zoid_api.h"
#include "zoid_mapping.h"
#include "zoid_protocol.h"

/* A structure we maintain in scratchpad, to exchange first packets read from
   the tree that were meant for the other CPU (in VN mode).  */
struct vn_shared
{
    int cpu; /* 0 for primary, 1 for secondary, -1 if not in use */
    char pad[12]; /* To guarantee quad-alignment for packet.  */
    struct ZoidSoftHeader softheader;
    char packet[TREE_DATA_SIZE];
} QUAD_ALIGN;

static int __zoid_vn_mode;
int __zoid_process_count;

static const char* mpi_mapping;

static void* stub_buffer = NULL;
static size_t stub_buffer_size = 0;

static int my_p2p_addr, ion_p2p_addr, my_pset_rank, pset_size;

/* 0 if this is the primary processor (in VN mode).  */
static int my_cpu_id;

/* Tree access mutexes for VN mode.  Used mainly for VC0, but occassionally
   (when MPI code is not running) also for VC1.  */
static struct BGL_Mutex *read_mutex, *write_mutex;

static void* other_process_window;
static unsigned int other_process_size;

static char packet[TREE_PACKET_SIZE] QUAD_ALIGN;

/* Only accessed if a process holds the read_mutex.  */
static volatile struct vn_shared* vn_shared;

static int last_errnum, last_excessive_size;

static int max_buffer_size;
static int ack_threshold;

static void __zoid_test_vc0(void) __attribute__((noinline));

/* Assert replacement.  We can't use ordinary assert here, because it
   wants to print to stderr, while we die in a middle of handling another
   command.  */
static void my_assert_fail(const char *assertion, const char *file,
			   unsigned int line, const char *function);
#define my_assert(expr) \
  ((expr) ? 0 : \
   (my_assert_fail(#expr, __FILE__, __LINE__, __PRETTY_FUNCTION__), 0))

/*
 * Zoid initialization function.  Runs before main().  Performs a handshake
 * with zoid on ION, exchanging information on job size, maximum buffer sizes,
 * etc.
 * FIXME: currently could run after other constructors.  We might want to move
 * it into __init_misc.
 */
void __init_zoid(void) __attribute__((constructor));
void
__init_zoid(void)
{
    BGLTreeStatusRegister status;
    BGLTreePacketHardHeader hardheader;
    BGLPersonality personality;
    const char* env;
    static int vn_sync;

    if (rts_get_personality(&personality, sizeof(personality)))
	abort();

    __zoid_vn_mode = BGLPersonality_virtualNodeMode(&personality);
    my_p2p_addr = BGLPersonality_treeP2PAddr(&personality);
    ion_p2p_addr = BGLPersonality_treeIONodeP2PAddr(&personality);
    my_pset_rank = BGLPersonality_rankInPset(&personality);
    pset_size = BGLPersonality_numNodesInPset(&personality);

    my_cpu_id = rts_get_processor_id();

    if ((env = getenv("BGLMPI_SIZE")))
    {
	if (!(__zoid_process_count = strtol(env, NULL, 10)))
	    abort();
    }
    else
    {
	/* This happens if invoking mpirun directly, without specifying
	   -np.  In that case the partition is completely full.  */
	__zoid_process_count = BGLPersonality_numComputeNodes(&personality)
	    * (__zoid_vn_mode ? 2 : 1);
    }

    /* User can specify alternative mapping.  It's a bit of a pain in the
       neck, as user-supplied data might be erroneous, and we don't really
       know how IBM code handles such errors...  */
    if (!(mpi_mapping = getenv("BGLMPI_MAPPING")))
	mpi_mapping = "XYZT";
    if (__zoid_mapping_init(mpi_mapping, __zoid_process_count, __zoid_vn_mode,
			    &personality))
	abort();

    if (__zoid_vn_mode && my_cpu_id == 0)
    {
	/* Check if CPU 1 is active.  If not, pretend that we are not in
	   VN mode, so as to avoid any overheads.  */

	unsigned rank;
	if (__zoid_mapping_to_rank(BGLPersonality_xCoord(&personality),
				   BGLPersonality_yCoord(&personality),
				   BGLPersonality_zCoord(&personality),
				   1, &rank) == -1)
	    __zoid_vn_mode = 0;
    }

    if (__zoid_vn_mode)
    {
	void* scratch_window;
	unsigned int scratch_size;

	if (rts_get_virtual_process_window(&other_process_window,
					   &other_process_size) ||
	    rts_get_scratchpad_window(&scratch_window, &scratch_size))
	{
	    abort();
	}

	/* We allocate from the top of the window, since msglayer seems to
	   be doing it from the bottom.  We can't use msglayer stuff here,
	   because msglayer might not be linked in, and would anyway not
	   yet be initialized at this point.  */
	vn_shared = scratch_window + scratch_size - sizeof(*vn_shared);

	if (my_cpu_id == 0)
	{
	    /* Primary CPU.  */
	    void* aligned;

	    vn_shared->cpu = -1;

	    if (!(read_mutex = rts_allocate_mutex()) ||
		!(write_mutex = rts_allocate_mutex()))
	    {
		abort();
	    }

	    aligned = (void*)(((unsigned int)&read_mutex) & ~0x1f);
	    rts_dcache_store(aligned, aligned + 0x20);
	    aligned = (void*)(((unsigned int)&write_mutex) & ~0x1f);
	    rts_dcache_store(aligned, aligned + 0x20);
	}
	else
	{
	    /* We are reading the address of the allocated mutex directly from
	       the memory of the primary process.  */
	    struct BGL_Mutex* volatile * primary_read_mutex_ptr =
		other_process_window + (unsigned int)&read_mutex;
	    struct BGL_Mutex* volatile * primary_write_mutex_ptr =
		other_process_window + (unsigned int)&write_mutex;

	    while (!*primary_read_mutex_ptr)
		;
	    read_mutex = *primary_read_mutex_ptr;

	    while (!*primary_write_mutex_ptr)
		;
	    write_mutex = *primary_write_mutex_ptr;
	}
    }

    __zoid_test_vc0();

    /* Send a wakeup notification to ZOID.  For VN mode, we make both
       processes send it.  Not sure if that's needed, but it won't hurt.  */
    BGLTreePacketHardHeader_InitP2P(&hardheader, PACKET_CLASS_CIO, 0,
				    ion_p2p_addr);
    /* Since this packet goes over VC1, not VC0, we don't bother with
       ZoidSoftHeader.  Instead, we put extended information about the
       process and job, to ease zoid's work.  */
    ((struct InitMsg*)packet)->pset_cpu_rank = my_pset_rank +
	my_cpu_id * pset_size;
    ((struct InitMsg*)packet)->p2p_addr = my_p2p_addr;
    ((struct InitMsg*)packet)->pid = getpid();
    ((struct InitMsg*)packet)->pset_rank = my_pset_rank;
    ((struct InitMsg*)packet)->cpu = my_cpu_id;
    /* We can't use __zoid_vn_mode below, because we could have reset it
       above.  */
    ((struct InitMsg*)packet)->vn_mode =
	BGLPersonality_virtualNodeMode(&personality);
    ((struct InitMsg*)packet)->total_proc = __zoid_process_count;
    strcpy(((struct InitMsg*)packet)->mapping, mpi_mapping);

    if (write_mutex)
	BGL_Mutex_Acquire(write_mutex);
    do
    {
	status = *(volatile BGLTreeStatusRegister*)BGL_MEM_TREE_VC1_STATUS0;
    } while (status.injpktcnt > 7);
    BGLTreeFIFO_send((BGLTreePacketHardHeader*)BGL_MEM_TREE_VC1_HDRIN,
		     (BGLQuad*)BGL_MEM_TREE_VC1_DATAIN,
		     &hardheader, (BGLQuad*)packet);
    if (write_mutex)
	BGL_Mutex_Release(write_mutex);

    /* Wait until ZOID suspends CIOD and sends us an ACK.  The ACKs are only
       sent once per each node, so in VN mode we use the primary process to
       receive them.  */
    if (my_cpu_id == 0)
    {
	/* Mutex not needed here; the other core is spinning.  */

	do
	{
	    status = *(volatile BGLTreeStatusRegister*)
		BGL_MEM_TREE_VC1_STATUS0;
	} while (status.recpktcnt == 0);

	BGLTreeFIFO_recv((BGLTreePacketHardHeader*)BGL_MEM_TREE_VC1_HDROUT,
			 (BGLQuad*)BGL_MEM_TREE_VC1_DATAOUT,
			 &hardheader, (BGLQuad*)packet);

	if (hardheader.p2p.pclass != PACKET_CLASS_CIO)
	    abort();

	max_buffer_size = ((struct InitMsgReply*)packet)->max_buffer_size;
	ack_threshold = ((struct InitMsgReply*)packet)->ack_threshold;

	if (__zoid_vn_mode)
	{
	    void* aligned;

	    aligned = (void*)(((unsigned int)&max_buffer_size) & ~0x1f);
	    rts_dcache_store(aligned, aligned + 0x20);
	    aligned = (void*)(((unsigned int)&ack_threshold) & ~0x1f);
	    rts_dcache_store(aligned, aligned + 0x20);

	    vn_sync = 1;
	    aligned = (void*)(((unsigned int)&vn_sync) & ~0x1f);
	    rts_dcache_store(aligned, aligned + 0x20);
	}
    }
    else
    {
	/* Virtual node, co-processor.  */
	volatile int* vn_sync_ptr = other_process_window +
	    (unsigned int)&vn_sync;
	volatile int* max_buffer_size_ptr = other_process_window +
	    (unsigned int)&max_buffer_size;
	volatile int* ack_threshold_ptr = other_process_window +
	    (unsigned int)&ack_threshold;

	/* Wait for primary process to receive an ACK.  */
	while (!*vn_sync_ptr)
	    ;

	max_buffer_size = *max_buffer_size_ptr;
	ack_threshold = *ack_threshold_ptr;
    }

    /* ZOID initialization is finished.  We can use I/O from this point on.  */
}

/*
 * Zoid termination function.  Runs from exit(), after all the atexit
 * functions.  Performs a final message exchange with zoid daemon to ensure
 * that there are no zoid packets in the network by the time the control
 * returns to the kernel.
 */
void
_cleanup(void)
{
    struct ZoidSoftHeader softheader, *softheader_p = &softheader;
    BGLTreeStatusRegister status;
    BGLTreePacketHardHeader hardheader;
    int command;
    static int vn_sync, vn_sync_2;

    /* Normally _IO_cleanup is a weak alias to _cleanup, but since we override
       _cleanup, we need to invoke the IO cleanup explicitly.  */
    extern void _IO_cleanup(void);
    _IO_cleanup();

    /* No more user-level zoid packets to exchange.
       Send terminating packet.  */
    BGLTreePacketHardHeader_InitP2P(&hardheader, PACKET_CLASS_CIO, 0,
				    ion_p2p_addr);
    softheader.zoid_id = ZOID_ID;
    softheader.pset_cpu_rank = my_pset_rank + my_cpu_id * pset_size;
    softheader.flags = ZOID_SOFTHEADER_FIRST_PACKET |
	ZOID_SOFTHEADER_LAST_PACKET;
    softheader.msg_length = sizeof(int);

    command = ZOID_TERMINATING_ID;
    memcpy(packet, &command, sizeof(command));

    if (write_mutex)
	BGL_Mutex_Acquire(write_mutex);
    do
    {
	status = *(volatile BGLTreeStatusRegister*)BGL_MEM_TREE_VC0_STATUS0;
    } while (status.injpktcnt > 7);
    BGLTreeFIFO_sendH((BGLTreePacketHardHeader*)BGL_MEM_TREE_VC0_HDRIN,
		      (BGLQuad*)BGL_MEM_TREE_VC0_DATAIN,
		      &hardheader, (BGLQuad*)softheader_p, (BGLQuad*)packet);
    if (write_mutex)
	BGL_Mutex_Release(write_mutex);

    if (__zoid_vn_mode)
    {
	/* Wait for the other process on the node to terminate.  This way
	   we avoid the problem of receiving messages which are not for us,
	   should the other process decide to communicate with zoid for a
	   little while longer.  */
	volatile int* vn_sync_ptr = other_process_window +
	    (unsigned int)&vn_sync;
	void* aligned;

	/* Change own status to "reached the barrier".  */
	vn_sync = 1;
	aligned = (void*)(((unsigned int)&vn_sync) & ~0x1f);
	rts_dcache_store(aligned, aligned + 0x20);

	/* Wait until the other process does the same on its side.  */
	while (!*vn_sync_ptr)
	    ;
    }

    /* At this point we are guaranteed that there are no packets traveling
       between this node and zoid, in either direction.  */
    if (my_cpu_id == 0)
    {
	void* aligned;
	/* Wait for an acknowledgement from zoid.  */

	/* No need to use the mutex anymore.  */
	do
	{
	    status = *(volatile BGLTreeStatusRegister*)
		BGL_MEM_TREE_VC0_STATUS0;
	} while (status.recpktcnt == 0);

	BGLTreeFIFO_recv((BGLTreePacketHardHeader*)BGL_MEM_TREE_VC0_HDROUT,
			 (BGLQuad*)BGL_MEM_TREE_VC0_DATAOUT,
			 &hardheader, (BGLQuad*)packet);
	/* Nothing useful is provided in the message.  */

	if (__zoid_vn_mode)
	{
	    /* Notify the other process that we are done.  */
	    vn_sync_2 = 1;
	    aligned = (void*)(((unsigned int)&vn_sync_2) & ~0x1f);
	    rts_dcache_store(aligned, aligned + 0x20);
	}
    }
    else
    {
	/* Wait for a notification from the primary process.  */
	volatile int* vn_sync_2_ptr = other_process_window +
	    (unsigned int)&vn_sync_2;

	while (!*vn_sync_2_ptr)
	    ;
    }

    /* Kiss-of-death exchanged; time to terminate.  */
}

/* This function is invoked from the stubs.  It sends the command over the
   collective/tree network to the I/O node, waits for a result and returns
   it.  It does not interpret either the command or the result in any way.

   Support for zero-copy makes things a bit more tricky, as some
   interpretation of the zero-copy arguments is required.  For example, if
   2-d array is passed as input, then the 1-d array with sizes is expected to
   be of size size_t*.  Results are interpreted for zero-copy results, because
   they contain embedded size information (see zoid_protocol.h for details).
*/
int
__zoid_submit_command(char* buffer, int cmd_len, int max_res_len, int userbuf,
		      const void* arr2d, const void* arr,
		      int arr_cnt, int arr_el_size, void* out_arr)
{
    BGLTreePacketHardHeader hardheader;
    struct ZoidSoftHeader softheader, *softheader_p = &softheader;
    char* in_buf_current;
    int in_len_remaining;
    int result_len = 0; /* To shut GCC up.  */
    int total_in_len;
    int pkt_count, ack_count;
    char *result_current = buffer;
    /* Reply always consists of at least one packet, so setting
       result_remaining to TREE_DATA_SIZE should be OK.  */
    int result_remaining = TREE_DATA_SIZE;
    /* The reason for result_remaining_extra is to adjust the length
       to be read by sizeof(int) for 2-D arrays, which store the length
       of the next segment at the end of each zero-copy data block.  Since
       initial block is not 0c, it is set to 0 for now.  */
    int result_remaining_extra = 0;
    /* These index the 2-D zero-copy results.  zeroc_max specifies the
       number of array lines, or -1 if 1-D result is being passed (a value
       of -2 indicates that the variable is not initialized yet).  zeroc_idx
       is simply an index from 0 to zeroc_max - 1. */
    int zeroc_idx = -1, zeroc_max = -2;
    int next_chunk_len;
    char* data_buf = result_current;

    my_assert(((unsigned int)buffer & 0xf) == 0);

    BGLTreePacketHardHeader_InitP2P(&hardheader, PACKET_CLASS_CIO, 0,
				    ion_p2p_addr);

    /* Calculate total size of zero-copy case.
       The command buffer always has enough space to round up to the full
       packet size (__zoid_allocate_buffer takes care of that).
       The zero-copy data always begins on the packet boundary for simplicity
       reasons.  In case of arr2d, each segment begins on the packet boundary.
    */
    total_in_len = cmd_len;
    cmd_len = TREE_BUFFER_ROUNDUP(cmd_len);
    if (!userbuf)
    {
	if (arr2d)
	{
	    int i;

	    for (i = 0; i < arr_cnt; i++)
	    {
		total_in_len = TREE_BUFFER_ROUNDUP(total_in_len);
		total_in_len += ((size_t*)arr)[i] * arr_el_size;
	    }
	}
	else if (arr)
	{
	    total_in_len = TREE_BUFFER_ROUNDUP(total_in_len);
	    total_in_len += arr_cnt * arr_el_size;
	}
    }

    if (total_in_len > max_buffer_size)
    {
	last_errnum = E2BIG;
	last_excessive_size = total_in_len - max_buffer_size;
	return -1;
    }

    /* Initialize the softheader for the first packet.  */
    softheader.zoid_id = ZOID_ID;
    softheader.pset_cpu_rank = my_pset_rank + my_cpu_id * pset_size;
    softheader.flags = ZOID_SOFTHEADER_FIRST_PACKET;
    softheader.msg_length = total_in_len;

    in_buf_current = buffer;
    in_len_remaining = cmd_len;

    if (userbuf)
	ack_count = in_len_remaining / TREE_DATA_SIZE - 1;
    else if (ack_threshold && total_in_len > ack_threshold * TREE_DATA_SIZE)
	ack_count = 0;
    else
	ack_count = -1;

    if (write_mutex)
	BGL_Mutex_Acquire(write_mutex);
    /* Send message.  */
    for (pkt_count = 0; in_len_remaining > 0; pkt_count++)
    {
	BGLTreeStatusRegister status;
	char* send_data_buf = in_buf_current;

	if (in_len_remaining < TREE_DATA_SIZE ||
	    (((unsigned int)in_buf_current) & 0xf))
	{
	    /* This code is only triggered for zero-copy data.  We copy to
	       an intermediate buffer if the size is less than the full packet,
	       or if data is not properly aligned.  */
	    memcpy(packet, in_buf_current, in_len_remaining < TREE_DATA_SIZE ?
		   in_len_remaining : TREE_DATA_SIZE);
	    send_data_buf = packet;
	}

	/* Is this the last packet?  */
	if (in_len_remaining <= TREE_DATA_SIZE &&
	    (arr2d ? zeroc_idx == arr_cnt - 1 : (!arr || zeroc_idx != -1)))
		softheader.flags |= ZOID_SOFTHEADER_LAST_PACKET;

	if (pkt_count == ack_count)
	    softheader.flags |= (userbuf ?
				 ZOID_SOFTHEADER_INPUT_USERBUF_PACKET :
				 ZOID_SOFTHEADER_NEED_ACK_PACKET);

	do
	{
	    status = *(volatile BGLTreeStatusRegister*)
		BGL_MEM_TREE_VC0_STATUS0;
	} while (status.injpktcnt > 7);
	BGLTreeFIFO_sendH((BGLTreePacketHardHeader*)BGL_MEM_TREE_VC0_HDRIN,
			  (BGLQuad*)BGL_MEM_TREE_VC0_DATAIN, &hardheader,
			  (BGLQuad*)softheader_p, (BGLQuad*)send_data_buf);

	in_len_remaining -= TREE_DATA_SIZE;
	in_buf_current += TREE_DATA_SIZE;

	if (in_len_remaining <= 0)
	{
	    if (arr2d)
	    {
		/* We make it a loop to properly handle a case when some
		   of the array entries would have zero length.  */
		for (zeroc_idx++; zeroc_idx < arr_cnt; zeroc_idx++)
		{
		    in_buf_current = ((void**)arr2d)[zeroc_idx];
		    in_len_remaining = ((size_t*)arr)[zeroc_idx] * arr_el_size;
		    if (in_len_remaining)
			break;
		}
	    }
	    else if (arr)
	    {
		if (++zeroc_idx == 0)
		{
		    in_buf_current = (char*)arr;
		    in_len_remaining = arr_cnt * arr_el_size;
		}
	    }
	}

	softheader.flags = 0;

	if (pkt_count == ack_count)
	{
	    struct ZoidSoftHeader ack_softheader,
		*ack_softheader_p = &ack_softheader;
	    /* We have to wait for an ACK from the server before we can send
	       the rest of the command.  */

	    if (write_mutex)
		BGL_Mutex_Release(write_mutex);

	    /* The code below is mostly a copy of part of the standard message
	       receiving code (further down), so look for explanations
	       there.  And, keep them in sync!  */
	    for (;;)
	    {
		BGLTreePacketHardHeader ack_hardheader;
		int cpu;

		if (read_mutex)
		    BGL_Mutex_Acquire(read_mutex);

		if (__zoid_vn_mode && vn_shared->cpu >= 0)
		{
		    if (vn_shared->cpu == my_cpu_id)
		    {
			ack_softheader = vn_shared->softheader;
			memcpy(packet, (void*)vn_shared->packet,
			       TREE_DATA_SIZE);

			vn_shared->cpu = -1;

			/* We are awaiting a single-packet reply, so we are
			   basically done.  */
			if (read_mutex)
			    BGL_Mutex_Release(read_mutex);
			break;
		    }
		    else
		    {
			if (read_mutex)
			    BGL_Mutex_Release(read_mutex);
			continue;
		    }
		}

		do
		{
		    status = *(volatile BGLTreeStatusRegister*)
			BGL_MEM_TREE_VC0_STATUS0;
		} while (status.recpktcnt == 0);
		BGLTreeFIFO_recvH((BGLTreePacketHardHeader*)
				  BGL_MEM_TREE_VC0_HDROUT,
				  (BGLQuad*)BGL_MEM_TREE_VC0_DATAOUT,
				  &ack_hardheader, (BGLQuad*)ack_softheader_p,
				  (BGLQuad*)packet);

		my_assert(ack_softheader.zoid_id == ZOID_ID);

		cpu = (ack_softheader.pset_cpu_rank - my_pset_rank) /
		    pset_size;
		if (cpu != my_cpu_id)
		{
		    vn_shared->cpu = cpu;
		    vn_shared->softheader = ack_softheader;
		    memcpy((void*)vn_shared->packet, packet, TREE_DATA_SIZE);
		}

		if (read_mutex)
		    BGL_Mutex_Release(read_mutex);

		/* Again, we are awaiting a single-packet reply, so if it's
		   for us, then we are done.  */
		if (cpu == my_cpu_id)
		    break;
	    }

	    /* ack_softheader and packet now contain the packet for us.
	       Actually, all we care about is the softheader.  */
	    my_assert(ack_softheader.flags & ZOID_SOFTHEADER_FIRST_PACKET);
	    my_assert(ack_softheader.flags & ZOID_SOFTHEADER_LAST_PACKET);

	    if (ack_softheader.errnum)
	    {
		/* The most likely reason is that user allocation callback
		   returned NULL.  We don't send anything more and return
		   prematurely.  */
		last_errnum = ack_softheader.errnum;

		if (last_errnum == E2BIG)
		{
		    my_assert(ack_softheader.msg_length == sizeof(int));
		    memcpy(&last_excessive_size, packet,
			   sizeof(last_excessive_size));
		}
		return -1;
	    }

	    my_assert(ack_softheader.flags & ZOID_SOFTHEADER_ACK_PACKET);

	    /* OK, ACK received -- resume sending!  */

	    if (write_mutex)
		BGL_Mutex_Acquire(write_mutex);
	} /* if (pkt_count == ack_count) */
    }

    if (write_mutex)
	BGL_Mutex_Release(write_mutex);

    /* Message is sent.  Now wait for a reply.
       There are two loops here.
       The outer one waits until we can enter the inner loop to receive a
       packet.  This is important in VN mode, where a packet for us might
       have been received by the other CPU; it is put in a special one-packet
       buffer visible from both CPUs then.
       The inner loop is the one that receives and handles packets.  It
       normally only terminates when a complete message has been received,
       or, in VN mode, when a packet for the other CPU has been encountered
       (see above).
       This is so complicated because we now allow message interleaving,
       where a stream of packets to one CPU can be interrupted by a stream of
       packets to the other CPU.  So we must be able to exit the inner loop,
       keeping all the state, and resume later.
       NOTE: part of this code is duplicated above, when waiting for an
       acknowledgement.  Make sure the codes are kept in synch!  */
    for (;;)
    {
	BGLTreeStatusRegister status;
	int already_received_1stpkt = 0;
	int cpu;

	/* We need to acquire and release the mutex in each iteration of the
	   outer loop, so that the other CPU can access the packet stored for
	   it in vn_shared.  */
	if (read_mutex)
	    BGL_Mutex_Acquire(read_mutex);

	if (__zoid_vn_mode && vn_shared->cpu >= 0)
	{
	    if (vn_shared->cpu == my_cpu_id)
	    {
		/* The other process has read a packet of a message
		   that was destined for me.  Copy this message into
		   appropriate structures.  "data_buf" is always correct
		   here, either through the initialization at the start
		   of this function, or through the code at the bottom
		   of the inner loop.  */
		softheader = vn_shared->softheader;
		memcpy(data_buf, (void*)vn_shared->packet, TREE_DATA_SIZE);

		/* Free the interprocess buffer.  */
		vn_shared->cpu = -1;

		already_received_1stpkt = 1;
	    }
	    else
	    {
		/* The message is not for me.  I can't read anything from the
		   tree, since the inter-process buffer is full, so if there
		   was a message in the network, and it was not for me, there
		   would be nowhere to store it.  */
		if (read_mutex)
		    BGL_Mutex_Release(read_mutex);
		continue;
	    }
	}
	/* Note: we are still holding the read mutex at this point.  */

	/* Inner loop starts.  */
	do
	{
	    if (!already_received_1stpkt)
	    {
		do
		{
		    status = *(volatile BGLTreeStatusRegister*)
			BGL_MEM_TREE_VC0_STATUS0;
		} while (status.recpktcnt == 0);
		BGLTreeFIFO_recvH((BGLTreePacketHardHeader*)
				  BGL_MEM_TREE_VC0_HDROUT,
				  (BGLQuad*)BGL_MEM_TREE_VC0_DATAOUT,
				  &hardheader, (BGLQuad*)softheader_p,
				  (BGLQuad*)data_buf);
	    }
	    else
		already_received_1stpkt = 0;

	    my_assert(softheader.zoid_id == ZOID_ID);

	    cpu = (softheader.pset_cpu_rank - my_pset_rank) / pset_size;
	    if (cpu != my_cpu_id)
	    {
		/* The packet just read is not for this CPU.
		   Place it in a shared area and break out of this loop
		   to let the other CPU handle it.  */
		vn_shared->cpu = cpu;
		vn_shared->softheader = softheader;
		memcpy((void*)vn_shared->packet, data_buf, TREE_DATA_SIZE);

		break;
	    }

	    if (softheader.flags & ZOID_SOFTHEADER_FIRST_PACKET)
	    {
		/* We have just received the first packet.  Initialize
		   everything.  */
		my_assert(data_buf == buffer);

		result_len = softheader.msg_length;
		last_errnum = softheader.errnum;

		if (last_errnum)
		{
		    /* This is only possible if there was an out-of-memory
		       condition, if the result would have been too large,
		       or if the backend was not loaded.  In either case, a
		       one-packet reply is sent.  */
		    my_assert(softheader.flags & ZOID_SOFTHEADER_LAST_PACKET);

		    if (last_errnum == E2BIG)
		    {
			/* In this case, the reply contains the excess size
			   value.  */
			my_assert(result_len == sizeof(int));
			memcpy(&last_excessive_size, buffer,
			       sizeof(last_excessive_size));
		    }

		    if (read_mutex)
			BGL_Mutex_Release(read_mutex);

		    return -1;
		}

		my_assert(result_len <= max_res_len);

		/* We always round up the buffer when allocating, so we take
		   that into consideration here, so as to avoid interim
		   buffers for the last packet of the main buffer (no such
		   luck with zero-copy result buffers...).  */
		result_remaining = TREE_BUFFER_ROUNDUP(result_len);
	    }

	    my_assert(softheader.msg_length == result_len);

	    if (data_buf != result_current)
	    {
		/* Because of buffer size limitations or alignment (that
		   only occur for 0c data), we had to use an auxiliary
		   buffer.  Now copy stuff to its intended destination.  */
		if (result_remaining > 0)
		    memcpy(result_current, data_buf, result_remaining <
			   TREE_DATA_SIZE ? result_remaining : TREE_DATA_SIZE);

		if (result_remaining < TREE_DATA_SIZE &&
		    result_remaining_extra)
		{
		    /* The just read packet contains (part of) the length
		       of the next chunk.  If we are lucky, it contains
		       the whole length (it's only 4 bytes...).  If we are
		       unlucky, then the length is split between two
		       packets.  */
		    if (result_remaining >= 0)
		    {
			if (result_remaining + result_remaining_extra <=
			    TREE_DATA_SIZE)
			{
			    /* A complete length is in this packet.  */
			    memcpy(&next_chunk_len,
				   data_buf + result_remaining,
				   sizeof(next_chunk_len));
			}
			else
			{
			    /* The beginning of length is in this
			       packet.  The next packet (handled right
			       below) will contain the rest.  */
			    memcpy(&next_chunk_len,
				   data_buf + result_remaining,
				   TREE_DATA_SIZE - result_remaining);
			}
		    }
		    else
		    {
			/* The end of length is in this packet.  The
			   beginning has already been read and stored.

			   Proof of correctness by example:
			   Imagine that when reading the beginning of
			   length, result_remaining was 237.  That packet
			   thus contained 240 - 237 = 3 bytes of length.
			   Thus, current packet contains 4 - 3 = 1
			   remaining byte, that we need to copy to the
			   byte starting at offset 3 of next_chunk_len.
			   result_remaining is now 237 - 240 = -3.  The
			   code below thus does what it is supposed to.  */
			memcpy(((char*)&next_chunk_len) - result_remaining,
			       data_buf,
			       result_remaining_extra + result_remaining);
		    }
		}
	    }

	    if (!(softheader.flags & ZOID_SOFTHEADER_LAST_PACKET))
	    {
		/* Prepare a buffer for the next packet of this message.
		   Note that that packet might be received outside this loop,
		   from the other CPU over the vn_shared buffer.  */

		result_current += TREE_DATA_SIZE;
		result_remaining -= TREE_DATA_SIZE;

		if (out_arr && result_remaining + result_remaining_extra <= 0)
		{
		    /* We have a zero-copy result, and have just finished
		       reading the previous chunk (be it the init chunk or
		       a 0c one).  */

		    if (zeroc_max == -2)
		    {
			/* We have not analyzed the 0c header yet, which means
			   that the chunk we just finished reading was the
			   init one.  */
			memcpy(&zeroc_max, buffer + result_len -
			       2 * sizeof(int), sizeof(zeroc_max));

			/* Copy the length of the first 0c chunk.  */
			memcpy(&result_remaining, buffer + result_len -
			       sizeof(int), sizeof(result_remaining));

			if (zeroc_max == -1)
			    /* 1-D array.  */
			    result_current = out_arr;
			else /* 2-D array.  */
			{
			    result_current = ((void**)out_arr)[0];
			    zeroc_idx = 0;

			    if (zeroc_idx < zeroc_max - 1)
				/* Account for the integer that specifies the
				   size of the next chunk.  */
				result_remaining_extra = sizeof(int);
			    else
				result_remaining_extra = 0;
			}
		    }
		    else
		    {
			/* The 0c header has already been analyzed.
			   We should never get here for 1-D arrays, or after
			   reading the last chunk of 2-D arrays.  */
			my_assert(zeroc_max != -1 &&
				  zeroc_idx < zeroc_max - 1);

			/* The length of the next chunk, that is stored at the
			   end of the just read one, is copied by code later
			   in the loop to next_chunk_len, because it cannot be
			   stored in user-supplied buffer (there is no space
			   there).  */
			result_remaining = next_chunk_len;

			zeroc_idx++;
			result_current = ((void**)out_arr)[zeroc_idx];

			if (zeroc_idx < zeroc_max - 1)
			    /* Account for the integer that specifies the size
			       of the next chunk.  */
			    result_remaining_extra = sizeof(int);
			else
			    result_remaining_extra = 0;
		    }
		}

		my_assert(result_remaining + result_remaining_extra > 0);

		/* Use auxiliary buffer if there is not enough space in the
		   current one or if it is unaligned.  Either condition is
		   only possible with 0c data chunks.  */
		if (result_remaining < TREE_DATA_SIZE ||
		    (((unsigned int)result_current) & 0xf))
		{
		    data_buf = packet;
		}
		else
		    data_buf = result_current;
	    }
	} while (!(softheader.flags & ZOID_SOFTHEADER_LAST_PACKET));

	if (read_mutex)
	    BGL_Mutex_Release(read_mutex);

	if (cpu == my_cpu_id &&
	    (softheader.flags & ZOID_SOFTHEADER_LAST_PACKET))
	{
	    break;
	}
    }

    return result_len;
}

void*
__zoid_allocate_buffer(size_t size)
{
    if (size > max_buffer_size)
    {
	/* This simply indicates an error on the caller's side.  We return
	   a hard error, which is propagated to the caller, who can then
	   adjust the request size and resubmit.  */
	last_errnum = E2BIG;
	last_excessive_size = size - max_buffer_size;
	return NULL;
    }

    if (size > stub_buffer_size)
    {
	free(stub_buffer);

	/* For efficiency reasons, we allocate no less than 4K, and always
	   a power-of-two, up to max_buffer_size.  */
	if (size < 4096)
	    size = 4096;
	else
	{
	    int i;

	    for (i = 4096; i < size; i <<= 1)
		;

	    if (i > max_buffer_size)
		i = max_buffer_size;

	    size = i;
	}

	if (!(stub_buffer = malloc(size)))
	{
	    last_errnum = ENOMEM;
	    stub_buffer_size = 0;
	}
	else
	    stub_buffer_size = size;
    }

    return stub_buffer;
}

void
__zoid_release_buffer(void* buffer)
{
    /* Currently, we don't free the buffer, but re-use it next time, if
       possible.  */
}

int
__zoid_error(void)
{
    return last_errnum;
}

/* If __zoid_error returns E2BIG, __zoid_excessive_size informs by how
   much the input was too large or the output would have been.  */
int
__zoid_excessive_size(void)
{
    return last_excessive_size;
}

static void
my_assert_fail(const char *assertion, const char *file, unsigned int line,
	       const char *function)
{
    struct ZoidSoftHeader softheader, *softheader_p = &softheader;
    BGLTreeStatusRegister status;
    BGLTreePacketHardHeader hardheader;
    static int *p = NULL;
    int n;

    /* Send an emergency packet with error description.  */
    BGLTreePacketHardHeader_InitP2P(&hardheader, PACKET_CLASS_CIO, 0,
				    ion_p2p_addr);
    softheader.zoid_id = ZOID_ID;
    softheader.pset_cpu_rank = my_pset_rank + my_cpu_id * pset_size;
    softheader.flags = ZOID_SOFTHEADER_FIRST_PACKET |
	ZOID_SOFTHEADER_LAST_PACKET | ZOID_SOFTHEADER_ASSERT_PACKET;

    n = snprintf(packet, TREE_DATA_SIZE, "%s:%u: %s%sAssertion `%s' failed!",
		 file, line, function ? function : "", function ? ": " : "",
		 assertion);
    if (n > TREE_DATA_SIZE)
	n = TREE_DATA_SIZE;

    softheader.msg_length = n;

    /* We are *dying* here.  No time to play with mutexes.  */
    do
    {
	status = *(volatile BGLTreeStatusRegister*)BGL_MEM_TREE_VC0_STATUS0;
    } while (status.injpktcnt > 7);
    BGLTreeFIFO_sendH((BGLTreePacketHardHeader*)BGL_MEM_TREE_VC0_HDRIN,
		      (BGLQuad*)BGL_MEM_TREE_VC0_DATAIN,
		      &hardheader, (BGLQuad*)softheader_p, (BGLQuad*)packet);

    /* Crash.  This has the advantage that control is immediately relinquished
       to CNK, which arranges for the core file to be generated.  */
    *p = 0;
}

/*
 * The only purpose of this function is to try to access VC0 early on, before
 * CIOD is suspended.  If a kernel with restricted access to VC0 is used,
 * this will give us a nice, easy to diagnose crash and a core file.
 */
static void
__zoid_test_vc0(void)
{
    BGLTreeStatusRegister status;

    status = *(volatile BGLTreeStatusRegister*)BGL_MEM_TREE_VC0_STATUS0;
}
