/****************************************************************************/
/* ZEPTOOS:zepto-info */
/*     This file is part of ZeptoOS: The Small Linux for Big Computers.
 *     See www.mcs.anl.gov/zeptoos for more information.
 */
/* ZEPTOOS:zepto-info */
/* */
/* ZEPTOOS:zepto-fillin */
/*     $Id: writev.c,v 1.6 2007/06/27 15:20:38 iskra Exp $
 *     ZeptoOS_Version: 1.2
 *     ZeptoOS_Heredity: FOSS_MOD
 *     ZeptoOS_License: GPL
 */
/* ZEPTOOS:zepto-fillin */
/* */
/* ZEPTOOS:zepto-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and UChicago Argonne, LLC.  2004, 2005, 2006, 2007
 *      ZeptoOS License: GPL
 * 
 *      This software is free.  See the file ZeptoOS/misc/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* ZEPTOOS:zepto-gpl */
/****************************************************************************/

#include <alloca.h>
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <sys/uio.h>

#include "unix.h"

//#include <stdio.h>

/* Note: most of the code below is duplicated in readv.c.  If anything needs
   to be changed in one of them, it probably also should be changed in the
   other!  */

ssize_t
__writev (int fd, const struct iovec *vector, int count)
{
    void** bases;
    size_t* lens;
    int i, min_i;
    static ssize_t max_writev_nbytes = -1;
    size_t bytes_total;
    ssize_t bytes_written;

    if (vector == NULL || count < 1)
    {
	__set_errno (EINVAL);
	return -1;
    }

    bases = alloca(sizeof(*bases) * count);
    lens = alloca(sizeof(*lens) * count);

    bytes_total = 0;
    for (i = 0; i < count; i++)
    {
	bases[i] = vector[i].iov_base;
	lens[i] = vector[i].iov_len;
	bytes_total += vector[i].iov_len;
    }

    bytes_written = 0;
    min_i = 0;
    do
    {
	ssize_t towrite, written;
	int error;
	int max_i;
	size_t max_i_len;

	towrite = 0;
	for (i = min_i; i < count; i++)
	    towrite += lens[i];

	if (max_writev_nbytes != -1 && towrite > max_writev_nbytes)
	{
	    ssize_t towrite_tmp;
#if 0
	    fprintf(stderr, "Adjusting writev of %d bytes from %d to %d\n",
		    bytes_total, towrite, max_writev_nbytes);
#endif
	    towrite = max_writev_nbytes;

	    towrite_tmp = 0;
	    for (max_i = min_i; max_i < count; max_i++)
	    {
		if (towrite_tmp + lens[max_i] >= towrite)
		    break;
		towrite_tmp += lens[max_i];
	    }

	    /* It should always exit the loop with the break.  */
	    assert(max_i < count);

	    /* max_i now points at the last array index to write with the
	       next call.  */

	    max_i_len = lens[max_i];
	    /* Decrease the length of the last element to write, so that
	       it does not exceed the buffer size.  */
	    lens[max_i] = towrite - towrite_tmp;
	}
	else
	{
	    max_i = count - 1;
	    max_i_len = lens[max_i];
	}
#if 0
	fprintf(stderr, "writev will be from count %d (offset %d, len %d) "
		"to %d (len %d)\n", min_i,
		bases[min_i] - vector[min_i].iov_base, lens[min_i],
		max_i, lens[max_i]);
#endif
	written = unix_writev(fd, (const void**)&bases[min_i], &lens[min_i],
			      max_i - min_i + 1);

	/* No matter if we succeeded or not, restore any changes we have
	   made to lens[max_i].  Also store the length as passed to writev
	   in max_i_len.  */
	{
	    size_t tmp;
	    tmp = max_i_len;
	    max_i_len = lens[max_i];
	    lens[max_i] = tmp;
	}

	if ((error = __zoid_error()))
	{
	    if (error != E2BIG)
	    {
		/* For a generic ZOID error, just bail out.  */
		__set_errno(error);
		return -1;
	    }
#if 0
	    fprintf(stderr, "Attempt to writev %d bytes returned excessive "
		    "size error by %d bytes\n", towrite,
		    __zoid_excessive_size());
#endif
	    /* We tried to send a too large writev request.  Adjust.
	       Note: this adjustment cannot be correct for 2-D array types,
	       because of full-packet alignment of each line.  Here, we don't
	       take this alignment into consideration.  As a consequence,
	       we will in some cases send less data than is actually possible,
	       but that shouldn't be a big deal.  */
	    max_writev_nbytes = towrite - __zoid_excessive_size();
	}
	else
	{
	    if (written < 0)
	    {
		/* For a generic writev() error, just bail out.  */
		__set_errno(-written);
		return -1;
	    }

	    bytes_written += written;

	    if (written != towrite)
		/* writev() as such succeeded, but it wrote fewer bytes than
		   expected.  We terminate prematurely then.  */
		break;

	    /* We wrote successfully.  Prepare for the next iteration, should
	       it be needed.  */
	    bases[max_i] += max_i_len;
	    lens[max_i] -= max_i_len;
	    if (lens[max_i] == 0)
		max_i++;
	    min_i = max_i;
	}
    } while (bytes_written < bytes_total);

    return bytes_written;
}

weak_alias (__writev, writev)
